from pathlib import Path

from app import create_app
from app.db.checkpoint import Database


# get the resources folder in the tests folder
RESOURCES = Path(__file__).parent / "resources"


def test_export_success():

    app = create_app()
    client = app.test_client()
    url = "/export"

    with open(RESOURCES / "sample_form.json") as fp:
        response = client.post(url, data={"drawflow_data": fp.read()})
        assert response.status_code == 200
        db = Database("library_name")
        db.load_history()
        db.clear_checkpoints()



"""
Driver function for the Flask application. This script implements the Flask
configurations, views and database functionality.
"""

from flask import Flask

from .routes import bp
from .demo import demo_bp


def create_app() -> Flask:
    """
    The Flask application factory function.
    """
    # create and configure the app
    app = Flask(__name__, instance_relative_config=True)
    app.static_folder = "static"
    app.url_map.strict_slashes = False
    app.config["UPLOAD_FOLDER"] = "db"
    app.register_blueprint(bp)
    app.register_blueprint(demo_bp)

    return app



import json
from pathlib import Path

from flask import Blueprint, Response, redirect, render_template, request, url_for
from werkzeug.utils import secure_filename

from .composition import CompositionParser
from .db.checkpoint import Database

bp = Blueprint("bp", __name__)


def valid_checkpoint(file_name: str) -> bool:

    fp = Path(file_name)
    return fp.is_file() and fp.suffix == ".json"


@bp.route("/")
def index() -> str:

    return render_template("index.html")


@bp.route("/canvas")
@bp.route("/canvas/<checkpoint>")
def canvas(checkpoint: str = None) -> str:

    element_divs = ""
    node_styles = ""
    df_box_divs = {}
    imported_drawflow = {}

    if checkpoint and valid_checkpoint(checkpoint):
        with open(checkpoint) as fp:
            checkpoint_data = json.loads(fp.read())
            if "drawflow" in checkpoint_data:
                imported_drawflow = {"drawflow": checkpoint_data["drawflow"]}
                element_divs = checkpoint_data["element_list_html"]
                df_box_divs = checkpoint_data["df_box_divs"]
                node_styles = checkpoint_data["node_styles_html"]

    return render_template(
        "canvas.html",
        element_divs=element_divs,
        df_box_divs=json.dumps(df_box_divs),
        node_styles=node_styles,
        imported_drawflow=imported_drawflow,
    )


@bp.route("/import", methods=["POST"])
def import_checkpoint() -> Response:

    checkpoint = request.files["checkpoint"]
    if valid_checkpoint(checkpoint.filename):
        file_name = secure_filename(checkpoint.filename)
        return redirect(url_for("canvas", checkpoint=file_name))

    return Response(status=500)


@bp.route("/export", methods=["POST"])
def export_checkpoint() -> Response:

    drawflow_data = json.loads(request.form["drawflow_data"])

    db = Database(drawflow_data["library"])
    db.load_history()
    db.save_checkpoint(drawflow_data)

    comp_parser = CompositionParser(drawflow_data["drawflow"], drawflow_data["library"])
    comp_parser.parse()
    comp_parser.generate_config()
    config_templ_str = render_template("run.py", **comp_parser.get_config())
    comp_parser.dump_config("run.py", config_templ_str)

    return Response(status=200)



"""
Represent a Drawflow node as a ComponentNode with data relevent for the
generated Python configuration file.

ComponentNode objects are intended to, but not restricted to, only work with
ComponentTree objects.
"""


class ComponentNode:
    """
    Structured representation of Drawflow data nodes.

    Attributes
    ----------
    class_name: str = None
        Name of the component class. e.g. an Adder SST component will require
        a single class for implementation. That same class may be reused with
        different parameter and link values. However, the configuration of the
        parameters and links will be consistent.

    type: int = 0
        Type of the Drawflow node, provided by the "id" field generated when a
        node is added to the canvas. In an unexpanded hierarchy structure, the
        type is a unique attribute useful for determining the number of copies
        of the node required.

    name: str = None
        Unique name of the component, generated by the class_name and the
        number of occurrences of the class_name in the entire hierarchy.

    parent: str = None
        Name of the parent node of the component in the hierarchy. At the first
        level, the parent of the nodes is "Home", an empty ComponentNode object.
        The nodes at every other level have parents that point to ComponentNode
        objects.

    links: list<dict> = None
        List of bidirectional links represented by the directed connections in
        the Drawflow canvas. A link is represented as followed:
            `{
                "from_port": str,
                "to_node_type": int,
                "to_port": str,
            }`,
        where:
        "from_port" is the name of the output connection of the current node,
        "to_node_type" is the type of node (Drawflow node "id") with the input
        connection, and
        "to_port" is the name of the input connection

    params: str = None
        Parameter values stored in the "data" field of the Drawflow nodes. A
        parameter is a dictionary that has been converted to an immutable
        string, since it does not serve a purpose in building the hierarchy.

    Public methods
    --------------
    Setters
    -------
    set_class_name(str)
    set_type(int)
    set_name(str)
    set_parent(str)
    set_links(list)
    set_params(str)

    Overloaded built-ins
    --------------------
    __eq__
    __hash__
    """

    def __init__(
        self,
        class_name: str = None,
        type: int = 0,
        name: str = None,
        parent: str = None,
        links: list = None,
        params: str = None,
    ) -> None:
        """
        Constructor for ComponentNode.

        Params
        ------
        class_name: str = None
        type: int = 0
        name: str = None
        parent: str = None
        links: list = None
        params: str = None

        Returns
        -------
        None
        """
        self.class_name = class_name
        self.type = type
        self.name = name
        self.parent = parent
        self.links = links
        self.params = params
        self.id = id(self)

    def set_class_name(self, class_name: str) -> None:
        """
        Sets class_name of object

        Params
        ------
        class_name: str = None

        Returns
        -------
        None
        """
        self.class_name = class_name

    def set_type(self, type: int) -> None:
        """
        Sets type of object

        Params
        ------
        type: int = 0

        Returns
        -------
        None
        """
        self.type = type

    def set_name(self, name: str) -> None:
        """
        Sets name of object

        Params
        ------
        name: str = None

        Returns
        -------
        None
        """
        self.name = name

    def set_parent(self, parent: str) -> None:
        """
        Sets parent name of object

        Params
        ------
        parent: str = None

        Returns
        -------
        None
        """
        self.parent = parent

    def set_links(self, links: list) -> None:
        """
        Sets links of object

        Params
        ------
        links: list = None

        Returns
        -------
        None
        """
        self.links = links

    def set_params(self, params: str) -> None:
        """
        Sets params of object

        Params
        ------
        params: str = None

        Returns
        -------
        None
        """
        self.params = params

    def __repr__(self) -> str:
        # debugging method
        return f"{self.class_name}{str(self.id)[-5:]}"

    def __eq__(self, other) -> bool:
        """Overloaded the equality operator for the object"""

        # if being compared to another ComponentNode object
        if isinstance(other, ComponentNode):
            # compare their class_name
            return self.class_name == other.class_name

        # if being compared to a string
        elif isinstance(other, str):
            # compare with the object's class_name
            return self.class_name == other

        # if being compared to an int
        elif isinstance(other, int):
            # compare with the object's id
            return self.id == other

        # if being compared to any other types
        raise TypeError(
            f"No methods implemented to check equality between 'ComponentNode' and {type(other)}"
        )

    def __hash__(self) -> int:
        """Overloaded the hash operator for the object"""
        return hash(self.id)



#!/usr/bin/env python
# -*- coding: utf-8 -*-

import sst

sst.setProgramOption("stopAtCycle", "5s")

fulladder78144 = sst.Component("fulladder78144", "library_name.fulladder")
fulladder78144.addParams({'clock': '1MHz', 'link_speed': '1ps'})
fulladder78192 = sst.Component("fulladder78192", "library_name.fulladder")
fulladder78192.addParams({'clock': '1MHz', 'link_speed': '1ps'})
fulladder78240 = sst.Component("fulladder78240", "library_name.fulladder")
fulladder78240.addParams({'clock': '1MHz', 'link_speed': '1ps'})
fulladder78288 = sst.Component("fulladder78288", "library_name.fulladder")
fulladder78288.addParams({'clock': '1MHz', 'link_speed': '1ps'})
fulladder78432 = sst.Component("fulladder78432", "library_name.fulladder")
fulladder78432.addParams({'clock': '1MHz', 'link_speed': '1ps'})
fulladder78480 = sst.Component("fulladder78480", "library_name.fulladder")
fulladder78480.addParams({'clock': '1MHz', 'link_speed': '1ps'})
fulladder78528 = sst.Component("fulladder78528", "library_name.fulladder")
fulladder78528.addParams({'clock': '1MHz', 'link_speed': '1ps'})
fulladder78576 = sst.Component("fulladder78576", "library_name.fulladder")
fulladder78576.addParams({'clock': '1MHz', 'link_speed': '1ps'})

sst.Link("fulladder78144-cout").connect(
    (fulladder78144, "cout"), (fulladder78432, "cin")
)
sst.Link("fulladder78144-cout").connect(
    (fulladder78144, "cout"), (fulladder78240, "cin")
)
sst.Link("fulladder78144-cout").connect(
    (fulladder78144, "cout"), (fulladder78192, "cin")
)
sst.Link("fulladder78192-cout").connect(
    (fulladder78192, "cout"), (fulladder78480, "cin")
)
sst.Link("fulladder78192-cout").connect(
    (fulladder78192, "cout"), (fulladder78288, "cin")
)
sst.Link("fulladder78240-cout").connect(
    (fulladder78240, "cout"), (fulladder78528, "cin")
)
sst.Link("fulladder78240-cout").connect(
    (fulladder78240, "cout"), (fulladder78288, "cin")
)
sst.Link("fulladder78288-cout").connect(
    (fulladder78288, "cout"), (fulladder78432, "opand2")
)
sst.Link("fulladder78432-cout").connect(
    (fulladder78432, "cout"), (fulladder78528, "cin")
)
sst.Link("fulladder78432-cout").connect(
    (fulladder78432, "cout"), (fulladder78480, "cin")
)
sst.Link("fulladder78480-cout").connect(
    (fulladder78480, "cout"), (fulladder78576, "cin")
)
sst.Link("fulladder78528-cout").connect(
    (fulladder78528, "cout"), (fulladder78576, "cin")
)


DF_BOX_DIVS = """
<div>
  <div class="title-box">
    <i class="fas fa-code"></i> {element}
  </div>

  <div class="modal micromodal-slide" id="modal-{element}" aria-hidden="true">
    <div class="modal__overlay" tabindex="-1" data-micromodal-close>
    <div id="modal-container" class="modal__container" role="dialog" aria-modal="true" aria-labelledby="modal-{element}-title">
      <header id="modal-header" class="modal__header">
      <h2 class="modal__title">
        {element}: {desc}
      </h2>
      <button type="button" class="modal__close" aria-label="Close modal" data-micromodal-close></button>
      </header>
      <div id="modal-content-content" class="modal-content-content">
      <div id="modal-content" class="modal__content">
        {input_tag}
      </div>
      <footer id="modal-footer" class="modal__footer">
        <button type="button" class="modal__btn" data-micromodal-close aria-label="Close this dialog window">Close</button>
      </footer>
      </div>
    </div>
    </div>
  </div>

  <div class="box dbclickbox" id="modal-{element}" ondblclick="showComponentModal(this.id)">
    {desc}
  </div>
</div>
"""

INPUT_TAG = """<input type="text" df-{key} placeholder="{key}"><br>"""

ELEMENT_DIV = """
<div class="drag-drawflow" draggable="true" ondragstart="drag(event)" data-node="{0}">
  <i class="fas fa-code"></i><span> {0}</span>
</div>
"""

NODE_INPUT_STYLE = """
.drawflow-node.{class_name} .inputs .input:nth-child({index}):before {{
  display: block;
  content: "{value}";
  position: relative;
  right: 120px;
}}
"""

NODE_OUTPUT_STYLE = """
.drawflow-node.{class_name} .outputs .output:nth-child({index}):before {{
  display: block;
  content: "{value}";
  position: relative;
  left: 30px;
}}
"""



"""
Represent the entire Drawflow node structure as a ComponentTree composed of
ComponentNode objects.

A ComponentTree is a tree of ComponentNode objects. Standalone, the class can
be instantiated by a raw Drawflow node structure composition data structure.
The object would however not be an accurate representation of the composition
if the Drawflow node structure contains submodules.
"""

from .node import ComponentNode


class ComponentTree:
    """
    Tree representation of Drawflow composition of nodes.

    Attributes
    ----------
    __composition: dict<ComponentNode, list<ComponentNode>> = None
        Prebuilt Drawflow composition data structure. In production, this
        parameter is always None due to the dynamic nature of the creation of
        the Drawflow compositions. The composition's keys are referred to as
        "modules" and its values as "nodes".

    root_key: str = "Home"
        The key of the root node in the composition data structure.

    __root: ComponentNode
        The root ComponentNode of the ComponentTree.

    __leaves: list<ComponentNode>
        List of leaf ComponentNode objects

    __tree: dict<ComponentNode, list<ComponentNode>>
        Tree representation of the Drawflow composition representing the true
        hierarchy of components.

    __height: int
        Height of the tree.

    Methods
    -------
    Public methods
    --------------
    add_parent(str)
    add_child(str, str, int, str, list, str)
    find_module(str)
    decompress()
    get_leaves()
    get_height()
    get_tree()

    Private methods
    ---------------
    __get_node_count(str)
    __get_children(ComponentNode)
    __decompress(ComponentNode)
    __get_leaves(dict, int)
    """

    def __init__(self, composition: dict = None, root_key: str = "Home") -> None:
        """
        Constructor for ComponentTree

        Params
        ------
        composition: dict<ComponentNode, list<ComponentNode>> = None
        root_key: str = "Home"

        Returns
        -------
        None
        """
        self.__composition = composition
        self.root_key = root_key

        # if a composition is provided
        if self.__composition:
            for module in self.__composition.keys():
                if module == self.root_key:
                    self.__root = module
                    break
        else:
            self.__composition = {}
            self.__root = None

        self.__leaves: list = []  # <list(ComponentNode)>
        self.__tree: dict = {}  # <dict(ComponentNode: list(ComponentNode))>
        self.__height: int = 0

    def add_parent(self, parent_name: str) -> None:
        """
        Add a ComponentNode as a parent in the hierarchy.

        The parent node is a ComponentNode object with only its class_name
        and name assigned as parent_name. The parent node is added to the
        composition as a key with an empty list as its value.

        Params
        ------
        parent_name: str
            Name of the parent node

        Returns
        -------
        None
        """
        module_node = ComponentNode(class_name=parent_name, name=parent_name)
        self.__composition[module_node] = []

    def add_child(
        self,
        parent_name: str,
        node_name: str,
        node_index: int,
        node_type: str,
        node_links: list,
        node_params: str,
    ) -> None:
        """
        Add a ComponentNode as a child in the hierarchy.

        Params
        ------
        parent_name: str
            Name (and class_name) of the parent ComponentNode

        node_name: str
            Name of the current ComponentNode

        node_index: int
            The index of the current ComponentNode's parent's children. The
            index does not signify a precedence or priority within the
            children nodes.

        node_type: str
            The type of the current ComponentNode.

        node_links: list
            The links of the current ComponentNode.

        node_params: str
            The params of the current ComponentNode.

        Returns
        -------
        None
        """
        module_node = self.find_module(parent_name)

        # append a new ComponentNode object with ComponentNode.class_name
        self.__composition[module_node].append(ComponentNode(class_name=node_name))
        node = self.__composition[module_node][node_index]

        node.set_parent(parent_name)
        node_count = self.__get_node_count(node_name)
        node.set_name(self.__get_node_name(node_name, node_count))
        node.set_type(node_type)
        node.set_links(node_links)
        node.set_params(node_params)

    def __get_node_name(self, node_name: str, count: int) -> str:
        """
        Format the node name with the ComponentNode's class_name and its
        occurrence in the composition.

        Params
        ------
        node_name: str
            Name of the current ComponentNode.

        count: int
            The occurrence of the current ComponentNode's class_name in the
            composition.

        Returns
        -------
        str: the formatted name for the current ComponentNode.
        """
        return f"{node_name}#{count}"

    def __get_node_count(self, node_class_name: str) -> int:
        """
        Count the occurrence of the current ComponentNode's class_name in the
        composition.

        Params
        ------
        node_class_name: str
            Name of the current ComponentNode.

        Returns
        -------
        count: int
            the occurrence of the current ComponentNode's class_name.
        """
        count = -1
        for module in self.__composition.keys():
            count += self.__composition[module].count(node_class_name)

        return count

    def find_module(self, node_name: str) -> ComponentNode:
        """
        Find a module in the composition (parent ComponentNode) by its name.

        Params
        ------
        node_name: str
            Name of the parent ComponentNode.

        Returns
        -------
        ComponentNode: the found parent ComponentNode or
        None: if the parent ComponentNode was not found.
        """
        for module in self.__composition.keys():
            if module == node_name:
                return module

    def __get_children(self, node: ComponentNode) -> list:
        """
        Generates new copies of children nodes from a module.

        Params
        ------
        node: str
            Parent ComponentNode.

        Returns
        -------
        list<ComponentNode>: list of newly generated ComponentNode children.
        """
        # iterating through the keys
        for module in self.__composition:
            # if the module key is found
            if module == node:
                # new and unique instances of ComponentNode objects with
                # attributes identical to the original copies
                return [
                    ComponentNode(
                        class_name=i.class_name,
                        type=i.type,
                        parent=i.parent,
                        name=i.name,
                        links=i.links,
                        params=i.params,
                    )
                    for i in self.__composition[module]
                ]

        # node is a leaf
        return []

    def __decompress(self, node: ComponentNode) -> dict:
        """
        Decompresses the flat hierarchical data structure by Drawflow.

        This recursive method converts the following data structure:
            {
                "root": ["a"],
                "a": ["b", "b"],
                "b": ["c", "c", "c"],
                "c": ["d"],
            }
        into the following:
            {"root": [
                {"a": [
                    {"b": [
                        {"c": [
                            {"d": []}
                        ]},
                        {"c": [
                            {"d": []}
                        ]},
                        {"c": [
                            {"d": []}
                        ]}
                    ]},
                    {"b": [
                        {"c": [
                            {"d": []}
                        ]},
                        {"c": [
                            {"d": []}
                        ]},
                        {"c": [
                            {"d": []}
                        ]}
                    ]},
                ]}
            ]}

        Params
        ------
        node: str
            Parent ComponentNode.

        Returns
        -------
        dict<ComponentNode, list<dict<ComponentNode>, list...>>:
            nested, decompressed version of the composition
        """
        return {node: [self.__decompress(n) for n in self.__get_children(node)]}

    def decompress(self) -> None:
        """
        Generates the root ComponentNode objects and initializes the values for
        the recursive decompress method to generate the tree.

        Params
        ------
        None

        Returns
        -------
        None
        """
        self.__root = self.find_module(self.root_key)
        self.__tree = self.__decompress(self.__root)

    def __get_leaves(self, subtree: dict, depth: int = 0) -> None:

        for key, value in subtree.items():
            if not value:
                self.__leaves.append(key)
                self.__height = max(self.__height, depth)

            for node in value:
                self.__get_leaves(node, depth + 1)

    def get_leaves(self) -> list:
        """
        Generates and returns the list of leaf ComponentNode objects.

        Params
        ------
        None

        Returns
        -------
        self.__leaves: list<ComponentNode>
            list of leaf ComponentNode objects.
        """
        self.__get_leaves(self.__tree)
        return self.__leaves

    def get_height(self) -> int:
        """
        Returns the height of the ComponentTree.

        Params
        ------
        None

        Returns
        -------
        self.__height: int
            the height of the ComponentTree.
        """
        return self.__height

    def get_tree(self) -> dict:
        """
        Returns the ComponentTree.

        Params
        ------
        None

        Returns
        -------
        self.__tree: dict<ComponentNode, list<dict<ComponentNode>, list...>>
            the ComponentTree.
        """
        return self.__tree



COMPONENT_INIT = """{name} = sst.Component("{name}", "{library}.{class_name}")"""
COMPONENT_PARAM = "{name}.addParams({params})"
COMPONENT_LINK = """sst.Link("{comp_out}-{link_out}").connect(
    ({comp_out}, "{link_out}"), ({comp_in}, "{link_in}")
)"""



"""
Resolves Drawflow connections between nodes to represent one-to-one
relationships between SST Links.

The class is a helper class and builds on an initialized ComponentTree object.
The connections between the nodes must be resolved for the Python configuration
file. Without the SST Links, the configuration file would instantiate SST
Components that do not have any functionality.
"""

from .node import ComponentNode


class Hierarchy:
    """
    Helper class to build SST Links from the ComponentNodes in an initialized
    ComponentTree object.

    Attributes
    ----------
    __tree: dict<ComponentNode, list<dict<ComponentNode>, list...>>
        the tree generated from a ComponentTree object.

    __hierarchy_links: list<tuple<str>>
        list of SST Links generated from the ComponentTree object's tree.

    Methods
    -------
    Public methods
    --------------
    get_path_to_root(ComponentNode, list)
    get_parent(ComponentNode)
    get_sibling_subtree(ComponentNode, int)
    resolve_from_port(ComponentNode, str)
    resolve_to_port(ComponentNode, int, str)
    parse_connection(str)
    resolve()
    get_links()

    Private methods
    ---------------
    __resolve_port(list, dict)
    __find_node_by_attr(dict, str, int)
    __get_parent(ComponentNode, dict)
    __resolve_hierarchy(dict)
    """

    def __init__(self, component_tree: dict) -> None:
        """
        Constructor for HierarchyResolver

        Params
        ------
        component_tree: dict<ComponentNode, list<dict<ComponentNode>, list...>>
            the tree generated from a ComponentTree object.

        Returns
        -------
        None
        """
        self.__tree = component_tree
        self.__hierarchy_links = []

    def __find_node_by_attr(self, subtree: dict, attr: str, data: int) -> tuple:
        """
        Recursively searches for a node by the specified attribute type and
        value.

        For example, to locate a node with a "type" attribute of 10 in the
        entire tree, run `__find_node_by_attr(tree, "type", 10)`.

        Params
        ------
        subtree: dict<ComponentNode, list<dict<ComponentNode>, list...>>
            A subtree of the tree to locate a node by the specified attribute.

        attr: str
            The attribute to locate the node with. The current attributes used
            in the program are: {"id", "type"}

        data: int
            The value of the attribute to locate the node with. The current
            attributes are both integer types.

        Returns
        -------
        key, subtree: tuple<ComponentNode, dict>
            a tuple of the found node and the subtree that was reduced from the
            recursive calls. If the node was not found, None is returned.
        """
        for key, value in subtree.items():
            if getattr(key, attr) == data:
                return key, subtree

            for node in value:
                found = self.__find_node_by_attr(node, attr, data)
                if found:
                    return found

    def get_path_to_root(self, node: ComponentNode, node_types_list: list) -> list:
        """
        Generates a list of node types to represent the shortest path between
        the specified node and the root.

        Params
        ------
        node: ComponentNode
            The specific node to find the shortest path to root.

        node_types_list: list<int>
            Initial list of node types between the specified node and the root.

        Returns
        -------
        list<int>: updated list of node types and the root.
        """
        # append the type of the current node to the list
        node_types_list.append(node.type)

        # append the type of the current node's parent to the list
        parent = self.get_parent(node)
        node_types_list.append(parent.type)

        # keep appending the types of the node's parents to the list until root
        # root.type is always 0
        while parent.type:
            parent = self.get_parent(parent)
            node_types_list.append(parent.type)

        return node_types_list

    def get_parent(self, node: ComponentNode) -> ComponentNode:
        """
        Initializes the recursive method to search for the parent of the
        current ComponentNode.

        Params
        ------
        node: ComponentNode
            The specific node to find the parent of.

        Returns
        -------
        ComponentNode: the parent of the current node or None if no parents
        are found.
        """
        return self.__get_parent(node, self.__tree)

    def __get_parent(self, node: ComponentNode, subtree: dict) -> ComponentNode:
        """
        Searches and returns the parent of the current ComponentNode.

        Params
        ------
        node: ComponentNode
            The specific node to find the parent of.

        subtree: dict<ComponentNode, list<dict<ComponentNode>, list...>>
            A subtree of the tree to locate the parent node.

        Returns
        -------
        key: ComponentNode
            parent node of the specified ComponentNode.
        """
        for key, value in subtree.items():
            for k in value:

                # if the id of the child node matches the current node, return
                # the parent
                if node.id == next(iter(k)).id:
                    return key

                parent = self.__get_parent(node, k)
                if parent:
                    return parent

    def get_sibling_subtree(self, node: ComponentNode, sibling_node_type: int) -> dict:

        current_node, _ = self.__find_node_by_attr(self.__tree, "id", node.id)

        current_parent = self.get_parent(current_node)
        current_parent, subtree = self.__find_node_by_attr(
            self.__tree, "id", current_parent.id
        )

        sibling_node = self.__find_node_by_attr(subtree, "type", sibling_node_type)
        while not sibling_node:
            current_parent = self.get_parent(current_parent)
            current_parent, subtree = self.__find_node_by_attr(
                self.__tree, "id", current_parent.id
            )
            sibling_node = self.__find_node_by_attr(subtree, "type", sibling_node_type)

        else:
            sibling_node, subtree = sibling_node

        return subtree

    def __resolve_port(self, node_types_list: list, subtree: dict) -> ComponentNode:

        while node_types_list:
            node, subtree = self.__find_node_by_attr(
                subtree, "type", node_types_list.pop()
            )
            if not subtree[node]:
                return node

    def resolve_from_port(self, node: ComponentNode, connection: str) -> tuple:

        connection_name, node_types_list = self.parse_connection(connection)

        if not node_types_list:
            return node, connection_name

        node_types_list = self.get_path_to_root(node, node_types_list)

        subtree = self.__tree
        return self.__resolve_port(node_types_list, subtree), connection_name

    def resolve_to_port(
        self, node: ComponentNode, to_node_type: int, connection: str
    ) -> tuple:

        subtree = self.get_sibling_subtree(node, to_node_type)
        connection_name, node_types_list = self.parse_connection(connection)
        if not node_types_list:
            return (
                self.__find_node_by_attr(subtree, "type", to_node_type)[0],
                connection_name,
            )

        return self.__resolve_port(node_types_list, subtree), connection_name

    def parse_connection(self, connection: str) -> tuple:
        """
        Parses connection string representing SST Links.

        The connection string is split on the node delimiter. The first node of
        the list is the name of the connection, and the rest of the list is its
        nested types.
        """
        connection_list = connection.split("#")
        return connection_list[0], [int(i) for i in connection_list[1:]]

    def resolve(self) -> None:

        self.__resolve_hierarchy(self.__tree)

    def __resolve_hierarchy(self, subtree: dict) -> None:

        for value in subtree.values():
            if not value:
                return
            for node in value:
                k = next(iter(node))
                for link in k.links:

                    from_node, from_port = self.resolve_from_port(k, link["from_port"])
                    to_node, to_port = self.resolve_to_port(
                        k,
                        link["to_node_type"],
                        link["to_port"],
                    )

                    self.__hierarchy_links.append(
                        (*(from_node, from_port), *(to_node, to_port))
                    )
                self.__resolve_hierarchy(node)

    def get_links(self) -> list:

        return self.__hierarchy_links



from datetime import datetime
import filecmp
import json
from pathlib import Path, PosixPath
from typing import Any


class Database:
    def __init__(self, library: str) -> None:

        self.checkpoints = Path().resolve() / ".checkpoints"
        self.library = library
        self.history_path = self.__get_path(f"{self.library}_log")
        self.history = []

    def __get_path(self, filename: str) -> PosixPath:

        return self.checkpoints / (filename + ".json")

    def __log_history(self) -> None:

        current_datetime = str(datetime.now())

        with open(self.history_path, "w") as fp:
            self.history.append(current_datetime)
            json.dump(self.history, fp)

    def __overwrite_history(self) -> None:

        with open(self.history_path, "w") as fp:
            self.history = self.history[:-2] + [self.history[-1]]
            json.dump(self.history, fp)

    def __get_checkpoint_name(self) -> PosixPath:

        return self.__get_path(f"{self.library}-{len(self.history) - 1}")

    def __get_last_checkpoint(self) -> PosixPath | None:

        if len(self.history):
            return self.__get_checkpoint_name()

    def load_history(self) -> None:

        if self.history_path.is_file():
            with open(self.history_path) as fp:
                self.history = json.loads(fp.read())

    def save_checkpoint(self, data: Any) -> None:

        last_checkpoint = self.__get_last_checkpoint()
        self.__log_history()
        new_checkpoint = self.__get_checkpoint_name()

        with open(new_checkpoint, "w") as fp:
            json.dump(data, fp)

        if last_checkpoint:
            if filecmp.cmp(last_checkpoint, new_checkpoint, shallow=False):
                self.__overwrite_history()
                new_checkpoint.unlink()

    def clear_checkpoints(self) -> None:

        for i in range(len(self.history)):
            self.__get_path(f"{self.library}-{i}").unlink(missing_ok=True)

        self.history_path.unlink(missing_ok=True)



import subprocess
import xml.etree.ElementTree as ET


class SSTInfoXMLParser:
    def __init__(self, file_name: str = "SSTInfo.xml") -> None:

        return_code = subprocess.run(["sst-info", "-no", file_name])
        self.tree = (
            ET.parse(file_name).getroot() if return_code.returncode == 0 else None
        )
        self.parsed_data = []
        self.children_keys = {"Port", "Parameter"}

    def xml_to_dict(self) -> list:

        for element in self.tree.findall("Element"):
            element_data = {
                "Name": element.attrib["Name"],
                "Component": [],
            }
            for component in element.findall("Component"):
                component_data = {
                    "Name": component.attrib["Name"],
                }

                for child_item in self.children_keys:
                    component_data[child_item] = []

                    for child_data in component.findall(child_item):
                        component_data[child_item].append(child_data.attrib)

                element_data["Component"].append(component_data)

            self.parsed_data.append(element_data)

    def get_dict(self):

        return self.parsed_data


if __name__ == "__main__":

    obj = SSTInfoXMLParser()
    obj.xml_to_dict()

    from pprint import pprint

    pprint(obj.get_dict())



import json

from flask import Blueprint, render_template

from .boilerplate.html import (
    DF_BOX_DIVS,
    ELEMENT_DIV,
    INPUT_TAG,
    NODE_INPUT_STYLE,
    NODE_OUTPUT_STYLE,
)


DEMO_COMPONENTS = [
    {
        "name": "div",
        "desc": "Division (SystemC)",
        "param": {
            "clock": "1MHz",
            "link_speed": "1ps",
        },
        "links": {
            "inputs": [
                "opand1",
                "opand2",
            ],
            "outputs": ["data_out"],
        },
    },
    {
        "name": "sumsq",
        "desc": "Sum of squares (SystemC)",
        "param": {
            "clock": "1MHz",
            "link_speed": "1ps",
        },
        "links": {
            "inputs": [
                "opand1",
                "opand2",
            ],
            "outputs": ["data_out"],
        },
    },
    {
        "name": "cacc",
        "desc": "Conditional accumulator (SystemC)",
        "param": {
            "clock": "1MHz",
            "link_speed": "1ps",
        },
        "links": {
            "inputs": [
                "dist",
                "inner",
                "outer",
            ],
            "outputs": ["new_inner", "new_outer"],
        },
    },
    {
        "name": "mt19937",
        "desc": "Mersenne Twister PRNG (Verilog)",
        "param": {
            "clock": "1MHz",
            "link_speed": "1ps",
        },
        "links": {
            "inputs": [
                "clk",
                "rst",
                "ready",
                "seed_val",
                "seed_start",
            ],
            "outputs": [
                "r_num",
                "valid",
                "busy",
            ],
        },
    },
    {
        "name": "swapprob",
        "desc": "Swap values on outcome (PyRTL)",
        "param": {
            "clock": "1MHz",
            "link_speed": "1ps",
        },
        "links": {
            "inputs": [
                "input_list",
                "prob",
            ],
            "outputs": [
                "new_list",
            ],
        },
    },
    {
        "name": "collision",
        "desc": "Coordinate overlap (PyRTL)",
        "param": {
            "clock": "1MHz",
            "link_speed": "1ps",
        },
        "links": {
            "inputs": [
                "h",
                "z",
                "d",
                "r",
            ],
            "outputs": [
                "new_h",
                "new_z",
                "new_d",
                "new_r",
            ],
        },
    },
    {
        "name": "gridwalk",
        "desc": "Walk grid subset (PyRTL)",
        "param": {
            "clock": "1MHz",
            "link_speed": "1ps",
        },
        "links": {
            "inputs": [
                "subset",
                "grid",
            ],
            "outputs": [
                "new_grid",
            ],
        },
    },
    {
        "name": "prob",
        "desc": "Probability test (SystemC)",
        "param": {
            "clock": "1MHz",
            "link_speed": "1ps",
        },
        "links": {
            "inputs": [
                "p",
            ],
            "outputs": [
                "data_out",
            ],
        },
    },
]


def __generate_drawflow(element_divs, df_box_divs, node_styles) -> tuple:

    for element in DEMO_COMPONENTS:
        input_tags = ""
        element_name = element["name"]

        for param in element["param"].keys():
            input_tags += INPUT_TAG.format(key=param)
        df_box_divs[element_name] = {}
        df_box_divs[element_name]["html"] = DF_BOX_DIVS.format(
            element=element_name, desc=element["desc"], input_tag=input_tags
        )
        df_box_divs[element_name]["links"] = element["links"]
        df_box_divs[element_name]["param"] = element["param"]
        node_styles += "\n".join(
            NODE_INPUT_STYLE.format(
                class_name=element_name, index=i + 1, value=j
            )
            for i, j in enumerate(element["links"]["inputs"])
        )
        node_styles += "\n".join(
            NODE_OUTPUT_STYLE.format(
                class_name=element_name, index=i + 1, value=j
            )
            for i, j in enumerate(element["links"]["outputs"])
        )

        element_divs += ELEMENT_DIV.format(element_name)

    return (
        element_divs,
        df_box_divs,
        node_styles,
    )


demo_bp = Blueprint("demo_bp", __name__)


@demo_bp.route("/demo")
def demo() -> str:

    element_divs = ""
    node_styles = ""
    df_box_divs = {}
    imported_drawflow = {}

    (element_divs, df_box_divs, node_styles) = __generate_drawflow(
        element_divs, df_box_divs, node_styles
    )

    return render_template(
        "canvas.html",
        element_divs=element_divs,
        df_box_divs=json.dumps(df_box_divs),
        node_styles=node_styles,
        imported_drawflow=imported_drawflow,
    )



import json

from .boilerplate.sst import (
    COMPONENT_INIT,
    COMPONENT_LINK,
    COMPONENT_PARAM,
)
from .component.tree import ComponentTree
from .component.hierarchy import Hierarchy


class CompositionParser:
    def __init__(self, data: dict, library: str) -> None:

        self.__raw_data = data
        self.__library = library
        self.__ctree = ComponentTree()
        self.__components_str_list = []
        self.__links_str_list = []
        self.__resolved_links = []

    def __copy_connections(self, component: dict, component_list: dict) -> None:

        output_names = component["data"]["links"]["outputs"]
        output_conns = component["outputs"].values()
        links = []
        for output_name, output_conn in zip(output_names, output_conns):
            for conn in output_conn["connections"]:
                to_node_type = conn["node"]
                to_node_conn_name = conn["output"]
                to_node_conn_num = int(to_node_conn_name[-1]) - 1
                to_node_port = component_list[to_node_type]["data"]["links"]["inputs"][
                    to_node_conn_num
                ]

                links.append(
                    {
                        "from_port": output_name,
                        "to_node_type": int(to_node_type),
                        "to_port": to_node_port,
                    }
                )

        return links

    def filter(self) -> None:

        for module_name, module in self.__raw_data.items():

            self.__ctree.add_parent(module_name)

            for component_list in module.values():

                for component_index, component in enumerate(component_list.values()):

                    # append a new ComponentNode object with ComponentNode.class_name
                    self.__ctree.add_child(
                        module_name,
                        component["name"],
                        component_index,
                        component["id"],
                        self.__copy_connections(component, component_list),
                        str(component["data"]["param"]),
                    )

    def generate_tree(self) -> ComponentTree:

        self.__ctree.decompress()
        return self.__ctree

    def dump_raw_data(self, file_name="dump.json") -> None:

        with open(file_name, "w") as dump_file:
            json.dump(self.__raw_data, dump_file, indent=4)

    def resolve_hierarchy(self) -> None:

        hr = Hierarchy(self.__ctree.get_tree())
        hr.resolve()
        self.__resolved_links = hr.get_links()

    def get_resolved_links(self) -> list:

        return self.__resolved_links

    def parse(self) -> None:

        self.filter()
        self.generate_tree()
        self.resolve_hierarchy()
        self.get_resolved_links()

    def generate_config(self) -> None:

        leaves = self.__ctree.get_leaves()
        for leaf in leaves:
            self.__components_str_list.append(
                COMPONENT_INIT.format(
                    name=leaf, library=self.__library, class_name=leaf.class_name
                )
            )
            self.__components_str_list.append(
                COMPONENT_PARAM.format(name=leaf, params=leaf.params)
            )

        self.__resolved_links = sorted(self.__resolved_links, key=lambda x: x[0].id)
        for link in self.__resolved_links:
            comp_out, link_out, comp_in, link_in = link
            self.__links_str_list.append(
                COMPONENT_LINK.format(
                    comp_out=comp_out,
                    link_out=link_out,
                    comp_in=comp_in,
                    link_in=link_in,
                )
            )

    def get_config(self) -> dict:

        return {
            "components": self.__components_str_list,
            "links": self.__links_str_list,
        }

    def dump_config(self, config_file_name, config_templ_str):

        with open(config_file_name, "w") as config_file:
            config_file.write(config_templ_str)



#!/usr/bin/env python
# -*- coasg: utf-8 -*-

import sst

sst.setProgramOption("stopAtCycle", "5s")
CLOCK = "1Hz"
LINK_DELAY = "1ps"

addersubtractor = sst.Component("addersubtractor", "calculator.addersubtractor")
addersubtractor.addParams(
    {"clock": CLOCK, "control": 1, "opand1": [1, 1, 0, 1], "opand2": [0, 1, 0, 1]}
)

ripplecarryadder = sst.Component("ripplecarryadder", "calculator.ripplecarryadder")
ripplecarryadder.addParams({"clock": CLOCK})

bintodec = sst.Component("bintodec", "calculator.bintodec")
bintodec.addParams({"clock": CLOCK})

full_add_0 = sst.Component("Full Adder 0", "calculator.fulladder")
full_add_0.addParams({"clock": CLOCK})

full_add_1 = sst.Component("Full Adder 1", "calculator.fulladder")
full_add_1.addParams({"clock": CLOCK})

full_add_2 = sst.Component("Full Adder 2", "calculator.fulladder")
full_add_2.addParams({"clock": CLOCK})

full_add_3 = sst.Component("Full Adder 3", "calculator.fulladder")
full_add_3.addParams({"clock": CLOCK})


sst.Link("b2d_sum_0").connect(
    (addersubtractor, "b2d_sum_0", LINK_DELAY), (bintodec, "sum_0", LINK_DELAY)
)
sst.Link("b2d_sum_1").connect(
    (addersubtractor, "b2d_sum_1", LINK_DELAY), (bintodec, "sum_1", LINK_DELAY)
)
sst.Link("b2d_sum_2").connect(
    (addersubtractor, "b2d_sum_2", LINK_DELAY), (bintodec, "sum_2", LINK_DELAY)
)
sst.Link("b2d_sum_3").connect(
    (addersubtractor, "b2d_sum_3", LINK_DELAY), (bintodec, "sum_3", LINK_DELAY)
)

sst.Link("as_opand1_0").connect(
    (addersubtractor, "as_opand1_0", LINK_DELAY), (ripplecarryadder, "as_opand1_0", LINK_DELAY)
)
sst.Link("as_opand1_1").connect(
    (addersubtractor, "as_opand1_1", LINK_DELAY), (ripplecarryadder, "as_opand1_1", LINK_DELAY)
)
sst.Link("as_opand1_2").connect(
    (addersubtractor, "as_opand1_2", LINK_DELAY), (ripplecarryadder, "as_opand1_2", LINK_DELAY)
)
sst.Link("as_opand1_3").connect(
    (addersubtractor, "as_opand1_3", LINK_DELAY), (ripplecarryadder, "as_opand1_3", LINK_DELAY)
)
sst.Link("as_opand2_0").connect(
    (addersubtractor, "as_opand2_0", LINK_DELAY), (ripplecarryadder, "as_opand2_0", LINK_DELAY)
)
sst.Link("as_opand2_1").connect(
    (addersubtractor, "as_opand2_1", LINK_DELAY), (ripplecarryadder, "as_opand2_1", LINK_DELAY)
)
sst.Link("as_opand2_2").connect(
    (addersubtractor, "as_opand2_2", LINK_DELAY), (ripplecarryadder, "as_opand2_2", LINK_DELAY)
)
sst.Link("as_opand2_3").connect(
    (addersubtractor, "as_opand2_3", LINK_DELAY), (ripplecarryadder, "as_opand2_3", LINK_DELAY)
)
sst.Link("as_sum_0").connect(
    (addersubtractor, "as_sum_0", LINK_DELAY), (ripplecarryadder, "as_sum_0", LINK_DELAY)
)
sst.Link("as_sum_1").connect(
    (addersubtractor, "as_sum_1", LINK_DELAY), (ripplecarryadder, "as_sum_1", LINK_DELAY)
)
sst.Link("as_sum_2").connect(
    (addersubtractor, "as_sum_2", LINK_DELAY), (ripplecarryadder, "as_sum_2", LINK_DELAY)
)
sst.Link("as_sum_3").connect(
    (addersubtractor, "as_sum_3", LINK_DELAY), (ripplecarryadder, "as_sum_3", LINK_DELAY)
)
sst.Link("as_cin_0").connect(
    (addersubtractor, "as_cin_0", LINK_DELAY), (ripplecarryadder, "as_cin_0", LINK_DELAY)
)
sst.Link("as_cout_3").connect(
    (addersubtractor, "as_cout_3", LINK_DELAY), (ripplecarryadder, "as_cout_3", LINK_DELAY)
)


sst.Link("add_opand1_0").connect(
    (full_add_0, "opand1", LINK_DELAY), (ripplecarryadder, "add_opand1_0", LINK_DELAY)
)
sst.Link("add_opand2_0").connect(
    (full_add_0, "opand2", LINK_DELAY), (ripplecarryadder, "add_opand2_0", LINK_DELAY)
)
sst.Link("add_cin_0").connect(
    (full_add_0, "cin", LINK_DELAY), (ripplecarryadder, "add_cin_0", LINK_DELAY)
)
sst.Link("add_sum_0").connect(
    (full_add_0, "sum", LINK_DELAY), (ripplecarryadder, "add_sum_0", LINK_DELAY)
)
sst.Link("add_cout_0").connect((full_add_0, "cout", LINK_DELAY), (full_add_1, "cin", LINK_DELAY))


sst.Link("add_opand1_1").connect(
    (full_add_1, "opand1", LINK_DELAY), (ripplecarryadder, "add_opand1_1", LINK_DELAY)
)
sst.Link("add_opand2_1").connect(
    (full_add_1, "opand2", LINK_DELAY), (ripplecarryadder, "add_opand2_1", LINK_DELAY)
)
sst.Link("add_sum_1").connect(
    (full_add_1, "sum", LINK_DELAY), (ripplecarryadder, "add_sum_1", LINK_DELAY)
)
sst.Link("add_cout_1").connect((full_add_1, "cout", LINK_DELAY), (full_add_2, "cin", LINK_DELAY))


sst.Link("add_opand1_2").connect(
    (full_add_2, "opand1", LINK_DELAY), (ripplecarryadder, "add_opand1_2", LINK_DELAY)
)
sst.Link("add_opand2_2").connect(
    (full_add_2, "opand2", LINK_DELAY), (ripplecarryadder, "add_opand2_2", LINK_DELAY)
)
sst.Link("add_sum_2").connect(
    (full_add_2, "sum", LINK_DELAY), (ripplecarryadder, "add_sum_2", LINK_DELAY)
)
sst.Link("add_cout_2").connect((full_add_2, "cout", LINK_DELAY), (full_add_3, "cin", LINK_DELAY))


sst.Link("add_opand1_3").connect(
    (full_add_3, "opand1", LINK_DELAY), (ripplecarryadder, "add_opand1_3", LINK_DELAY)
)
sst.Link("add_opand2_3").connect(
    (full_add_3, "opand2", LINK_DELAY), (ripplecarryadder, "add_opand2_3", LINK_DELAY)
)
sst.Link("add_sum_3").connect(
    (full_add_3, "sum", LINK_DELAY), (ripplecarryadder, "add_sum_3", LINK_DELAY)
)
sst.Link("add_cout_3").connect(
    (full_add_3, "cout", LINK_DELAY), (ripplecarryadder, "add_cout_3", LINK_DELAY)
)
